export class ForceUnmuteContainerV2Plugin {

  constructor() {
    this.name = 'Force-Unmute-Container-V2-Plugin';
    this.client = null;
  }

  unmute(container) {
    // if this game is using springroll.Sound, ensure that it's unmuted
    const localSpringroll = container.iframe.contentWindow.springroll;
    if (localSpringroll && localSpringroll.Sound && localSpringroll.Sound.instance) {
      localSpringroll.Sound.instance.systemMuted = false;
    }
  };

  /**
   *
   * @param {Object} params param object
   * @param {BellHop} params.client Bellhop client
   * @memberof SpringrollForceUnmuteContainerV2Plugin
   * @return {Promise} resolved if client exists
   */
  async preload({ client }) {
    this.client = client;
  }

  /**
   * @param {Container} container SpringrollContainer instance
   * @memberof SpringrollForceUnmuteContainerV2Plugin
   * @return {undefined}
   */
  init(container) {
    this.client.on('opened', () => {
      this.unmute(container);
    });

    this.client.on('open', () => {
      // specifically listen for the loadDone event that SpringRoll 0.3 games emit so they also get unmuted
      this.client.on('loadDone', () => {
        this.unmute(container);
      });
    });
  }
}
