const Signal = /** @class */ (function ()
{
    function Signal()
    {
        this.listeners = [];
    }
    /**
     * Adds a listener that is only called once before it is removed
     */
    Signal.prototype.addOnce = function (callback, context)
    {
        this.listeners.push({
            once: true,
            callback,
            context,
        });
    };
    /**
     * Adds a listener for this signal
     */
    Signal.prototype.add = function (callback, context)
    {
        this.listeners.push({
            once: false,
            callback,
            context,
        });
    };
    /**
     * Dispatches a message for this signal
     */
    Signal.prototype.dispatch = function (arg1, arg2, arg3)
    {
        const _this = this;

        for (let _i = 0, _a = this.listeners; _i < _a.length; _i++)
        {
            const listener = _a[_i];

            listener.callback.call(listener.context, arg1, arg2, arg3);
        }
        this.listeners
            .filter(function (listener)
            {
                return listener.once;
            })
            .forEach(function (listener)
            {
                return _this.remove(listener.callback, listener.context);
            });
    };
    /**
     * Removes a listener from the set of listeners
     */
    Signal.prototype.remove = function (callback, context)
    {
        this.listeners = this.listeners.filter(function (listener)
        {
            return listener.context !== context || listener.callback !== callback;
        });
    };

    return Signal;
})();

/* ! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = function (d, b)
{
    extendStatics = Object.setPrototypeOf
        || ({ __proto__: [] } instanceof Array && function (d, b)
        {
            d.__proto__ = b;
        })
        || function (d, b)
        {
            for (const p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
        };

    return extendStatics(d, b);
};

function __extends(d, b)
{
    extendStatics(d, b);
    function __()
    {
        this.constructor = d;
    }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

var __assign = function ()
{
    __assign = Object.assign || function __assign(t)
    {
        for (var s, i = 1, n = arguments.length; i < n; i++)
        {
            s = arguments[i];
            for (const p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }

        return t;
    };

    return __assign.apply(this, arguments);
};

function __awaiter(thisArg, _arguments, P, generator)
{
    function adopt(value)
    {
        return value instanceof P ? value : new P(function (resolve)
        {
            resolve(value);
        });
    }

    return new (P || (P = Promise))(function (resolve, reject)
    {
        function fulfilled(value)
        {
            try
            {
                step(generator.next(value));
            }
            catch (e)
            {
                reject(e);
            }
        }
        function rejected(value)
        {
            try
            {
                step(generator.throw(value));
            }
            catch (e)
            {
                reject(e);
            }
        }
        function step(result)
        {
            result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected);
        }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body)
{
    let _ = { label: 0, sent()
    {
        if (t[0] & 1) throw t[1];

        return t[1];
    }, trys: [], ops: [] };
    let f;
    let y;
    let t;
    let g;

    return g = { next: verb(0), throw: verb(1), return: verb(2) }, typeof Symbol === 'function' && (g[Symbol.iterator] = function ()
    {
        return this;
    }), g;
    function verb(n)
    {
        return function (v)
        {
            return step([n, v]);
        };
    }
    function step(op)
    {
        if (f) throw new TypeError('Generator is already executing.');
        while (_)
        {
            try
            {
                if (f = 1, y && (t = op[0] & 2 ? y.return : op[0] ? y.throw || ((t = y.return) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
                if (y = 0, t) op = [op[0] & 2, t.value];
                switch (op[0])
                {
                    case 0: case 1: t = op; break;
                    case 4: _.label++;

                        return { value: op[1], done: false };
                    case 5: _.label++; y = op[1]; op = [0]; continue;
                    case 7: op = _.ops.pop(); _.trys.pop(); continue;
                    default:
                        if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2))
                        {
                            _ = 0; continue;
                        }
                        if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3])))
                        {
                            _.label = op[1]; break;
                        }
                        if (op[0] === 6 && _.label < t[1])
                        {
                            _.label = t[1]; t = op; break;
                        }
                        if (t && _.label < t[2])
                        {
                            _.label = t[2]; _.ops.push(op); break;
                        }
                        if (t[2]) _.ops.pop();
                        _.trys.pop(); continue;
                }
                op = body.call(thisArg, _);
            }
            catch (e)
            {
                op = [6, e]; y = 0;
            }
            finally
            {
                f = t = 0;
            }
        }
        if (op[0] & 5) throw op[1];

        return { value: op[0] ? op[1] : void 0, done: true };
    }
}

function hasCordova()
{
    try
    {
        return !!window.cordova;
    }
    catch (e)
    {
        // TODO: Finish rewiring the logging in the DLM to use a proper logging class, w/ enable/disable capabilities
        console.log('Not inside of Cordova environment.');

        return false;
    }
}

const Storage = /** @class */ (function ()
{
    function Storage(location)
    {
        this.isWriting = false;
        this.savingPromise = null;
        this.savingResolve = null;
        this.dataIsDirty = false;
        this.data = {};
        this.location = location;
    }
    Storage.prototype.save = function (data)
    {
        return __awaiter(this, void 0, void 0, function ()
        {
            let dataToWrite;
            const _this = this;

            return __generator(this, function (_a)
            {
                switch (_a.label)
                {
                    case 0:
                        this.data = data || this.data;
                        if (!this.isWriting) return [3 /* break*/, 1];
                        this.dataIsDirty = true;
                        if (this.savingPromise)
                        {
                            return [2 /* return*/, this.savingPromise];
                        }
                        this.savingPromise = new Promise(function (resolve)
                        {
                            _this.savingResolve = resolve;
                        });

                        return [2 /* return*/, this.savingPromise];
                    case 1:
                        dataToWrite = this.data;
                        this.isWriting = true;

                        return [4 /* yield*/, this.writeToDisk(dataToWrite)];
                    case 2:
                        _a.sent();
                        this.isWriting = false;
                        if (!this.dataIsDirty) return [3 /* break*/, 4];
                        this.dataIsDirty = false;

                        return [4 /* yield*/, this.save(this.data)];
                    case 3:
                        _a.sent();
                        if (this.savingResolve)
                        {
                            this.savingResolve();
                        }
                        this.savingPromise = null;
                        this.savingResolve = null;
                        _a.label = 4;
                    case 4: return [2];
                }
            });
        });
    };
    Storage.prototype.load = function ()
    {
        return this.readFromDisk();
    };
    Storage.prototype.writeToDisk = function (data)
    {
        const _this = this;

        return new Promise(function (resolve, reject)
        {
            if (hasCordova())
            {
                window.resolveLocalFileSystemURL(cordova.file.dataDirectory, function (dirEntry)
                {
                    dirEntry.getFile(_this.location, { create: true }, function (fileEntry)
                    {
                        fileEntry.createWriter(function (writer)
                        {
                            writer.onwriteend = function ()
                            {
                                resolve();
                            };
                            const writeObject = data;

                            writer.write(JSON.stringify(writeObject));
                        }, reject);
                    }, reject);
                }, reject);
            }
            else
            {
                return true;
            }
        });
    };
    Storage.prototype.readFromDisk = function ()
    {
        const _this = this;

        if (this.isWriting)
        {
            return Promise.resolve(this.data);
        }

        return new Promise(function (resolve, reject)
        {
            if (hasCordova())
            {
                window.resolveLocalFileSystemURL(cordova.file.dataDirectory, function (dirEntry)
                {
                    dirEntry.getFile(_this.location, { create: true }, function (fileEntry)
                    {
                        fileEntry.file(function (file)
                        {
                            const reader = new FileReader();

                            reader.onloadend = function (e)
                            {
                                let json = null;

                                try
                                {
                                    if (!e.target.result)
                                    {
                                        resolve(null);
                                    }
                                    else
                                    {
                                        json = JSON.parse(e.target.result);
                                        resolve(json);
                                    }
                                }
                                catch (err)
                                {
                                    reject(err);
                                }
                            };
                            reader.onerror = function (evt)
                            {
                                reject(evt);
                            };
                            reader.readAsText(file);
                        }, reject);
                    });
                }, reject);
            }
            else
            {
                resolve(_this.data);
            }
        });
    };
    Storage.prototype.clear = function ()
    {
        const _this = this;

        return new Promise(function (resolve, reject)
        {
            window.resolveLocalFileSystemURL(cordova.file.dataDirectory, function (dir)
            {
                dir.getFile(_this.location, { create: false }, function (fileEntry)
                {
                    fileEntry.remove(resolve, reject, resolve);
                });
            });
        });
    };

    return Storage;
})();

const State = /** @class */ (function ()
{
    function State(id, instant, _progressWeight)
    {
        if (id === void 0)
        {
            id = 'no-id';
        }
        if (instant === void 0)
        {
            instant = false;
        }
        if (_progressWeight === void 0)
        {
            _progressWeight = 1;
        }
        this.signals = {
            onComplete: new Signal(),
            onProgress: new Signal(),
            onCancel: new Signal(),
            onError: new Signal(),
        };
        this.state = State.PENDING;
        this.id = id;
        this.instant = instant;
        this._progressWeight = _progressWeight;
        if (this.instant)
        {
            this._progressWeight = 0;
        }
        this._progress = 0;
    }
    Object.defineProperty(State.prototype, 'progressWeight', {
        get()
        {
            return this._progressWeight;
        },
        enumerable: true,
        configurable: true,
    });
    State.prototype.begin = function ()
    {
        console.warn('The `begin()` method needs to be overridden by a child class.');
    };
    State.prototype.cancel = function (cb)
    {
        cb();
    };
    State.prototype._begin = function ()
    {
        if (this.state === State.PENDING)
        {
            if (!this.instant)
            {
                this.state = State.ACTIVE;
                this.signals.onProgress.dispatch(0);
                // resume a State
                if (this.begin)
                {
                    this.begin();
                }
                else
                {
                    throw new Error('a non instant state requires a begin function!');
                }
            }
            else
            {
                this.complete();
            }
        }
    };
    State.prototype._cancel = function (cb)
    {
        if (!cb)
        {
            throw new Error('No callback provided for state cancel');
        }
        if (this.state === State.ACTIVE)
        {
            this.state = State.PENDING;
            this.cancel(cb);
        }
        else
        {
            cb();
        }
    };
    State.prototype.complete = function ()
    {
        this.state = State.COMPLETE;
        this.updateProgress(1);
        this.signals.onComplete.dispatch();
    };
    Object.defineProperty(State.prototype, 'progress', {
        /**
         * return a progres value between 0 and 1
         */
        get()
        {
            return this._progress;
        },
        enumerable: true,
        configurable: true,
    });
    /**
     * this sets the internal progress value and also sends out the progress signal
     * @param value the current progress
     */
    State.prototype.updateProgress = function (value)
    {
        this._progress = value;
        this.signals.onProgress.dispatch(1);
    };
    State.PENDING = 0;
    State.ACTIVE = 1;
    State.COMPLETE = 2;

    return State;
})();

const Cleaning = /** @class */ (function (_super)
{
    __extends(Cleaning, _super);
    function Cleaning(fileToDelete, id, progressWeight)
    {
        if (id === void 0)
        {
            id = 'CLEANING';
        }
        if (progressWeight === void 0)
        {
            progressWeight = 1;
        }
        const _this = _super.call(this, id, false, progressWeight) || this;

        _this.fileToDelete = fileToDelete;

        return _this;
    }
    Cleaning.prototype.begin = function ()
    {
        const _this = this;

        window.resolveLocalFileSystemURL(this.fileToDelete, function (fileEntry)
        {
            console.log('CLEANING', fileEntry);
            if (!fileEntry.isFile)
            {
                fileEntry.getDirectory('', { create: false }, function (unpackDir)
                {
                    console.log('Delete Folder Done : ', unpackDir);
                    unpackDir.removeRecursively(function ()
                    {
                        console.log('clean: folder deleted');
                        _this.complete();
                    }, function ()
                    {
                        console.log('clean: folder delete error');
                        _this.complete();
                    });
                }, function ()
                {
                    console.log('clean: folder not there..');
                });
            }
            else
            {
                fileEntry.remove(function ()
                {
                    console.log('clean: complete');
                    _this.complete();
                    // The file has been removed successfully
                }, function (error)
                {
                    console.log('clean: fail ', error);
                    _this.signals.onError.dispatch(_this, error);
                    // Error deleting the file
                }, function ()
                {
                    console.log('clean: file not there..');
                    // The file doesn't exist
                    _this.complete();
                });
            }
        }, function ()
        {
            console.log('clean: file not found..');
            _this.complete();
        });
    };

    return Cleaning;
})(State);

function getDataDirectory()
{
    return new Promise(function (resolve, reject)
    {
        window.resolveLocalFileSystemURL(window.cordova.file.dataDirectory, resolve, reject);
    });
}
function createNextDirectory(fileSystem, folderNames, index)
{
    return __awaiter(this, void 0, void 0, function ()
    {
        const _this = this;

        return __generator(this, function (_a)
        {
            return [2 /* return*/, new Promise(function (resolve, reject)
            {
                fileSystem.getDirectory(folderNames[index], { create: true }, function (nextFileSystem)
                {
                    return __awaiter(_this, void 0, void 0, function ()
                    {
                        return __generator(this, function (_a)
                        {
                            switch (_a.label)
                            {
                                case 0:
                                    index++;
                                    if (!(index === folderNames.length)) return [3 /* break*/, 1];
                                    // done!
                                    resolve(nextFileSystem);

                                    return [3 /* break*/, 3];
                                case 1: return [4 /* yield*/, createNextDirectory(nextFileSystem, folderNames, index)];
                                case 2:
                                    nextFileSystem = _a.sent();
                                    resolve(nextFileSystem);
                                    _a.label = 3;
                                case 3: return [2];
                            }
                        });
                    });
                }, reject);
            })];
        });
    });
}
function getDirectory(directoryPath, create)
{
    if (create === void 0)
    {
        create = true;
    }

    return __awaiter(this, void 0, void 0, function ()
    {
        let splits;
        let fileSystem;
        let directory;
        let newDirectory;

        return __generator(this, function (_a)
        {
            switch (_a.label)
            {
                case 0:
                    splits = directoryPath.split('/');
                    if (splits[0] === '.')
                    {
                        splits.shift();
                    }
                    if (splits[splits.length - 1].includes('.'))
                    {
                        splits.pop();
                    }

                    return [4 /* yield*/, getDataDirectory()];
                case 1:
                    fileSystem = (_a.sent());

                    return [4 /* yield*/, new Promise(function (resolve)
                    {
                        fileSystem.getDirectory(directoryPath, { create: false }, function (dir)
                        {
                            resolve(dir);
                        }, function ()
                        {
                            resolve(null);
                        });
                    })];
                case 2:
                    directory = _a.sent();
                    if (directory)
                    {
                        return [2 /* return*/, directory];
                    }
                    if (!create) return [3 /* break*/, 4];

                    return [4 /* yield*/, createNextDirectory(fileSystem, splits, 0)];
                case 3:
                    newDirectory = _a.sent();

                    return [2 /* return*/, newDirectory];
                case 4: throw new Error('[cordova get directory] directory does not exist');
            }
        });
    });
}

function copyFile(fileName, destination)
{
    return __awaiter(this, void 0, void 0, function ()
    {
        return __generator(this, function (_a)
        {
            return [2 /* return*/, new Promise(function (resolve, reject)
            {
                const error = function (e)
                {
                    console.log(`[cordova copy] copy fail: ${fileName} to ${destination}`);
                    reject(e);
                };

                window.resolveLocalFileSystemURL(fileName, function (fileEntry)
                {
                    const relativeDestination = destination.replace(cordova.file.dataDirectory, './');
                    const fileName = relativeDestination.split('/').pop();

                    getDirectory(relativeDestination)
                        .then(function (directory)
                        {
                            fileEntry.copyTo(directory, fileName, function ()
                            {
                                console.log(`[cordova copy] copy success :${fileName} to ${destination}`);
                                resolve();
                            }, error);
                        });
                }, error);
            })];
        });
    });
}

const Copy = /** @class */ (function (_super)
{
    __extends(Copy, _super);
    function Copy(target, destination, id, progressWeight)
    {
        if (id === void 0)
        {
            id = 'COPYING';
        }
        if (progressWeight === void 0)
        {
            progressWeight = 1;
        }
        const _this = _super.call(this, id, false, progressWeight) || this;

        _this.target = target;
        _this.destination = destination;

        return _this;
    }
    Copy.prototype.begin = function ()
    {
        const _this = this;

        copyFile(this.target, this.destination).then(function ()
        {
            _this.complete();
        })
            .catch(function (e)
            {
                console.log(e);
            });
    };
    Copy.prototype.cancel = function (cb)
    {
        cb();
    };

    return Copy;
})(State);

const Downloading = /** @class */ (function (_super)
{
    __extends(Downloading, _super);
    function Downloading(url, destination, id, progressWeight)
    {
        if (id === void 0)
        {
            id = 'DOWNLOADING';
        }
        if (progressWeight === void 0)
        {
            progressWeight = 1;
        }
        const _this = _super.call(this, id, false, progressWeight) || this;

        _this.url = url;
        _this.destination = destination;
        _this.transfer = null;
        _this._request = null;

        return _this;
    }
    Downloading.prototype.begin = function ()
    {
        const _this = this;
        const relativeDestination = this.destination.replace(cordova.file.dataDirectory, './');
        const fileName = relativeDestination.split('/').pop();

        getDirectory(relativeDestination).then(function (fs)
        {
            fs.getFile(fileName, { create: true, exclusive: false }, function (fileEntry)
            {
                const request = _this._request = new XMLHttpRequest();

                request.onprogress = function (event)
                {
                    _this.updateProgress(event.loaded / event.total);
                };
                request.open('GET', _this.url, true);
                request.responseType = 'blob';
                request.onload = function (oEvent)
                {
                    const blob = request.response;

                    if (blob)
                    {
                        fileEntry.createWriter(function (writer)
                        {
                            writer.onwrite = function ()
                            {
                                console.log('[download] write complete!? ');
                                _this.complete();
                            };
                            writer.write(blob);
                        });
                    }
                    else
                    {
                        throw new Error('[download] we didnt get an XHR response!');
                    }
                };
                request.send(null);
            }, function (err)
            {
                throw new Error('[download] error getting file! ');
            });
            console.log('[download] ', fs);
        })
            .catch(function (e)
            {
                console.log('[download] fail', e);
            });
    };
    Downloading.prototype.cancel = function (cb)
    {
        if (this._request)
        {
            this._request.abort();
        }
        cb();
    };

    return Downloading;
})(State);

const DownloadingCordova = /** @class */ (function (_super)
{
    __extends(DownloadingCordova, _super);
    function DownloadingCordova(url, destination, id, progressWeight)
    {
        if (id === void 0)
        {
            id = 'DOWNLOADING';
        }
        if (progressWeight === void 0)
        {
            progressWeight = 1;
        }
        const _this = _super.call(this, id, false, progressWeight) || this;

        _this.url = url;
        _this.destination = destination;
        _this.transfer = null;
        _this._didCancel = false;

        return _this;
    }
    DownloadingCordova.prototype.begin = function ()
    {
        const _this = this;
        const transfer = this.transfer = new FileTransfer();

        transfer.onprogress = function (event)
        {
            if (event.lengthComputable)
            {
                _this.updateProgress(event.loaded / event.total);
                // TODO - an event here!
                // console.log(event.loaded / event.total);
            }
        };
        transfer.download(this.url, this.destination, function (entry)
        {
            console.log(`download complete: ${entry.toURL()}`);
            _this.complete();
        }, function (error)
        {
            // expecting this if we manually cancelled!
            if (_this._didCancel)
            {
                return;
            }
            _this.signals.onError.dispatch(_this, error);
            console.log(error);
            console.log(`download error source ${error.source}`);
            console.log(`download error target ${error.target}`);
            console.log(`download error code ${error.code}`);
        }, false, {});
    };
    DownloadingCordova.prototype.cancel = function (cb)
    {
        this._didCancel = true;
        const transfer = this.transfer;

        console.log('ABORTING');
        if (transfer)
        {
            transfer.abort();
        }
        cb();
    };

    return DownloadingCordova;
})(State);

const Done = /** @class */ (function (_super)
{
    __extends(Done, _super);
    function Done()
    {
        return _super.call(this, 'done', true) || this;
    }

    return Done;
})(State);

const Waiting = /** @class */ (function (_super)
{
    __extends(Waiting, _super);
    function Waiting()
    {
        return _super.call(this, 'WAITING', true) || this;
    }

    return Waiting;
})(State);

const Unzipping = /** @class */ (function (_super)
{
    __extends(Unzipping, _super);
    function Unzipping(zipLocation, destination, progressWeight)
    {
        if (progressWeight === void 0)
        {
            progressWeight = 1;
        }
        const _this = _super.call(this, 'UNZIPPING', false, progressWeight) || this;

        _this.zipLocation = zipLocation;
        _this.destination = destination;
        _this.canceled = false;

        return _this;
    }
    Unzipping.prototype.begin = function ()
    {
        const _this = this;

        this.canceled = false;
        console.log('unzipping.. ');
        console.log(this.zipLocation);
        window.zip.unzip(this.zipLocation, this.destination, function (result)
        {
            if (result === 0)
            {
                if (!_this.canceled)
                {
                    console.log(result);
                    console.log('unzip complete!');
                    _this.complete();
                }
            }
            else if (result === 1 || result === -1)
            {
                console.log('unzip error!');
                _this.signals.onError.dispatch(_this, new Error(`zip failed error code ${result}`));
            }
            else
            {
                console.log('fail..');
            }
        }, function (progress)
        {
            // console.log("progress:", progress)
            _this.updateProgress(progress.loaded / progress.total);
        });
    };
    Unzipping.prototype.cancel = function (cb)
    {
        this.canceled = true;
        window.zip.stop(cb);
    };

    return Unzipping;
})(State);

let count = 0;
const SleepState = /** @class */ (function (_super)
{
    __extends(SleepState, _super);
    function SleepState(id, delay)
    {
        if (delay === void 0)
        {
            delay = 200;
        }
        const _this = _super.call(this, id || (`TEST ${count++}`), false) || this;

        _this.delay = delay;
        _this.progressInterval = 0;
        _this.counter = 0;

        return _this;
    }
    SleepState.prototype.begin = function ()
    {
        const _this = this;

        this.counter = 0;
        this.progressInterval = setInterval(function ()
        {
            _this.counter += 100;
            if (_this.counter > _this.delay)
            {
                clearInterval(_this.progressInterval);
                _this.complete();
            }
            else
            {
                _this.signals.onProgress.dispatch(_this.counter / _this.delay);
            }
        }, 100);
    };
    SleepState.prototype.cancel = function (cb)
    {
        if (this.progressInterval)
        {
            clearInterval(this.progressInterval);
        }
        cb();
    };

    return SleepState;
})(State);

const DeletedState = /** @class */ (function (_super)
{
    __extends(DeletedState, _super);
    function DeletedState()
    {
        const _this = _super.call(this, 'deleted', true) || this;

        _this.error = null;

        return _this;
    }

    return DeletedState;
})(State);

const ErrorState = /** @class */ (function (_super)
{
    __extends(ErrorState, _super);
    function ErrorState()
    {
        const _this = _super.call(this, 'error', true) || this;

        _this.error = null;

        return _this;
    }

    return ErrorState;
})(State);

const Flow = /** @class */ (function ()
{
    function Flow()
    {
        this.signals = {
            onResume: new Signal(),
            onProgress: new Signal(),
            onError: new Signal(),
            onCancel: new Signal(),
            onStateChange: new Signal(),
            onComplete: new Signal(),
        };
        this.active = false;
        this.currentState = null;
        this.currentPosition = 0;
        this.states = [];
        this.errorState = new ErrorState();
        this.deletedState = new DeletedState();
    }
    Flow.prototype.getStateId = function ()
    {
        if (!this.currentState)
        {
            throw new Error('current state is null');
        }

        return this.currentState.id;
    };
    Flow.prototype.add = function (state)
    {
        this.states.push(state);
        // first state!
        if (!this.currentState)
        {
            this.currentState = this.states[0];
        }

        return this;
    };
    Flow.prototype.getIndex = function ()
    {
        return this.currentPosition;
    };
    Flow.prototype.set = function (index)
    {
        this.currentPosition = index;
        if (index === -1)
        {
            this.currentState = this.errorState;
        }
        else
        {
            this.currentState = this.states[index];
        }
    };
    Flow.prototype.resume = function ()
    {
        let _a;

        if ((this.currentPosition === -1))
        {
            this.set(0);
        }
        this.active = true;
        if (((_a = this.currentState) === null || _a === void 0 ? void 0 : _a.state) !== State.PENDING)
        {
            // already processing..
            return;
        }
        this.next();
    };
    Object.defineProperty(Flow.prototype, 'finished', {
        /**
         * @return {boolean} true if the download has finished
         */
        get()
        {
            return this.getIndex() >= this.states.length - 1;
        },
        enumerable: true,
        configurable: true,
    });
    Flow.prototype.next = function ()
    {
        if (!this.currentState)
        {
            throw new Error('current state is null');
        }
        // console.log("NEXT>", this, this.currentState, this.currentState.state)
        this.currentState.signals.onComplete.remove(this.next, this);
        this.currentState.signals.onError.remove(this.onError, this);
        this.currentState.signals.onProgress.remove(this.onProgress, this);
        if (this.currentState.state !== State.PENDING)
        {
            this.currentPosition++;
        }
        if (this.currentPosition < this.states.length)
        {
            // resume a Queue
            this.currentState = this.states[this.currentPosition];
            if (!this.currentState)
            {
                throw new Error('current state is null');
            }
            this.currentState.signals.onComplete.addOnce(this.next, this);
            this.currentState.signals.onError.addOnce(this.onError, this);
            this.currentState.signals.onProgress.add(this.onProgress, this);
            this.signals.onStateChange.dispatch(this);
            this.currentState._begin();
            // console.log(this.currentState);
        }
        else
        {
            console.log('FINISHED FLOW');
            this.currentPosition = this.states.length - 1;
            this.signals.onComplete.dispatch();
        }
    };
    Flow.prototype.onError = function (state, error)
    {
        this.errorState.error = error;
        this.currentPosition = -1;
        this.currentState = this.errorState;
        this.active = false;
        this.signals.onError.dispatch(this);
    };
    Object.defineProperty(Flow.prototype, 'progress', {
        get()
        {
            let prog = 0;
            let total = 0;

            if (this.finished)
            {
                return 1;
            }
            for (let i = 0; i < this.states.length; i++)
            {
                const state = this.states[i];
                const progressWeight = state.progressWeight;

                if (i < this.currentPosition)
                {
                    // complete..
                    prog += Number(progressWeight);
                    total += Number(progressWeight);
                }
                else if (i === this.currentPosition)
                {
                    // active..
                    prog += state.progress * progressWeight;
                    total += Number(progressWeight);
                }
                else
                {
                    // not yet started..
                    // active..
                    total += Number(progressWeight);
                }
            }

            return prog / total;
        },
        enumerable: true,
        configurable: true,
    });
    Flow.prototype.onProgress = function (progress)
    {
        this.signals.onProgress.dispatch(this.progress);
    };
    Flow.prototype.pause = function ()
    {
        const _this = this;

        return new Promise(function (resolve)
        {
            if (!_this.active)
            {
                resolve();
            }
            else
            {
                // go back to previous state...
                // cancel a Queue
                _this.active = false;
                if (_this.currentState)
                {
                    _this.currentState.signals.onComplete.remove(_this.next, _this);
                    _this.currentState._cancel(resolve);
                }
                else
                {
                    resolve();
                }
            }
        });
    };
    Flow.prototype.getSave = function ()
    {
        return { cp: this.currentPosition };
    };
    Flow.prototype.delete = function ()
    {
        return __awaiter(this, void 0, void 0, function ()
        {
            return __generator(this, function (_a)
            {
                switch (_a.label)
                {
                    case 0: return [4 /* yield*/, this.pause()];
                    case 1:
                        _a.sent();
                        this.currentPosition = -1;
                        this.currentState = new DeletedState();
                        this.signals.onStateChange.dispatch(this);

                        return [2];
                }
            });
        });
    };

    return Flow;
})();

const CalculateSize = /** @class */ (function (_super)
{
    __extends(CalculateSize, _super);
    function CalculateSize(folder, id, progressWeight)
    {
        if (id === void 0)
        {
            id = 'CALCULATE_SIZE';
        }
        if (progressWeight === void 0)
        {
            progressWeight = 1;
        }
        const _this = _super.call(this, id, false, progressWeight) || this;

        _this.folder = folder;
        _this.transfer = null;
        _this.canceled = false;
        _this.cancelCallback = undefined;
        _this.size = 0;
        _this.loaded = 0;

        return _this;
    }
    CalculateSize.prototype.begin = function ()
    {
        const _this = this;

        this.loaded = 0;
        console.log('begin size check..', this.folder);
        window.resolveLocalFileSystemURL(this.folder, function (dirEntry)
        {
            _this.cdvDirectorySize(dirEntry, function (size)
            {
                _this.size = size;
                console.log('calculating size..');
                _this.updateProgress(1);
                console.log(`file size calculation complete: ${size}`);
                _this.complete();
            }, 1);
        });
    };
    CalculateSize.prototype.cdvDirectorySize = function (dirEntry, cbDone, ratio)
    {
        const _this = this;
        const reader = dirEntry.createReader();
        let sizeResult = 0;

        reader.readEntries(function (entries)
        {
            const segSize = ratio / entries.length;
            var processNext = function ()
            {
                // has it been canceled?
                if (_this.canceled)
                {
                    // lets stop and do a callback!
                    if (_this.cancelCallback)
                    {
                        _this.cancelCallback();
                    }

                    return;
                }
                if (entries.length)
                {
                    const entry = entries.pop();

                    if (entry instanceof FileEntry)
                    {
                        entry.getMetadata(function (metadata)
                        {
                            sizeResult += metadata.size;
                            _this.loaded += segSize;
                            _this.updateProgress(_this.loaded);
                            processNext();
                        }, function (error)
                        {
                            processNext();
                        });
                    }
                    else
                    {
                        _this.cdvDirectorySize(entry, function (size)
                        {
                            sizeResult += size;
                            processNext();
                        }, segSize);
                    }
                }
                else
                {
                    cbDone(sizeResult);
                }
            };

            processNext();
        }, function (error)
        {
            // doesn't handle errors recursively i know..
            // but probably better to fail through incorrect size than fatally.
            cbDone(sizeResult);
        });
    };
    CalculateSize.prototype.cancel = function (cb)
    {
        this.canceled = true;
        this.cancelCallback = cb;
    };

    return CalculateSize;
})(State);

/**
 * A Download represents a zip being downloaded.
 *
 * internally the download process has the following states:
 *
 * Waiting - waiting for download to begin
 *
 * Downloading - the actual download of the zip file
 *
 * Unzipping - unzipping the downloaded file to the out put directory
 *
 * Cleaning - deleting the downloaded zip
 *
 * Calculating Size - calculating the size of the game that has been downloaded
 *
 * Done - game is ready to go!
 *
 */
const Download = /** @class */ (function (_super)
{
    __extends(Download, _super);
    function Download(gameData, basePath, preferCordova)
    {
        if (preferCordova === void 0)
        {
            preferCordova = false;
        }
        const _this = _super.call(this) || this;

        _this.data = gameData;
        _this.basePath = basePath;
        const zipDestination = `${_this.basePath}zips/${_this.data.downloadID}/${_this.data.versionID}.zip`;
        const gameDestination = _this.unpackPath;
        // reset state array before building Flow

        _this.states = [];
        const bundled = !(_this.data.archivePath.includes('http://') || _this.data.archivePath.includes('https://'));

        if (hasCordova())
        {
            var DownloadAction = null;
            // unzipping does not work when the zip is in the www directory on android
            // so if the game is bundled, we copy it to the to the download location
            // rather than downloading there.

            if (bundled)
            {
                DownloadAction = Copy;
            }
            else if (preferCordova)
            {
                DownloadAction = DownloadingCordova;
            }
            else
            {
                DownloadAction = Downloading;
            }
            _this.add(new Waiting())
                .add(new DownloadAction(_this.data.archivePath, zipDestination))
                .add(new Unzipping(zipDestination, gameDestination))
                .add(new Cleaning(zipDestination, undefined, 0.01));
            // sometimes the unpacked filesize will be provided by the CMS
            // this saves us a fairly intensive step of calculating the file size!
            // speeds up 'downloads' loads (100 - 200 % faster) on low end devices!
            if (!_this.data.unpackedSize)
            {
                const calculateSize_1 = new CalculateSize(gameDestination);

                calculateSize_1.signals.onComplete.add(function ()
                {
                    _this.data.unpackedSize = calculateSize_1.size;
                });
                _this.add(calculateSize_1);
            }
            _this.add(new Done());
        }
        else
        {
            var DownloadAction = bundled ? 'COPYING' : 'DOWNLOADING';

            _this.add(new Waiting())
                .add(new SleepState(DownloadAction, 2000))
                .add(new SleepState('UNZIPPING', 1000))
                .add(new SleepState('CLEANING', 200))
                .add(new SleepState('CALCULATE_SIZE', 2000))
                .add(new Done());
        }
        if (_this.data.flowIndex !== undefined)
        {
            _this.set(_this.data.flowIndex);
        }

        return _this;
    }
    Object.defineProperty(Download.prototype, 'unpackSize', {
        /**
         * Returns the unpacked size of the game in bytes
         * @return {number} the file size of the game once unpacked
         */
        get()
        {
            const unpackSize = this.data.unpackedSize;

            if (!unpackSize)
            {
                // is this the right behavior? has to be a number the way this is written....
                return -1;
            }

            return unpackSize;
        },
        enumerable: true,
        configurable: true,
    });
    /**
     * a handy Promise that resolves when a download is complete or has already finished
     * resolves when a download is complete
     * @return {Promise} a promise that is resolved when the download is complete
     */
    Download.prototype.onReady = function ()
    {
        const _this = this;

        if (this.finished)
        {
            return Promise.resolve(this);
        }

        return new Promise(function (resolve)
        {
            _this.signals.onComplete.addOnce(resolve);
        });
    };
    Object.defineProperty(Download.prototype, 'archivePath', {
        get()
        {
            return this.data.archivePath;
        },
        enumerable: true,
        configurable: true,
    });
    Object.defineProperty(Download.prototype, 'unpackPath', {
        get()
        {
            return `${this.basePath}downloads/${this.data.downloadID}/${this.data.versionID}`;
        },
        enumerable: true,
        configurable: true,
    });
    Object.defineProperty(Download.prototype, 'downloadID', {
        get()
        {
            return this.data.downloadID;
        },
        enumerable: true,
        configurable: true,
    });
    Object.defineProperty(Download.prototype, 'updated', {
        get()
        {
            return this.data.updated || -1;
        },
        enumerable: true,
        configurable: true,
    });
    Object.defineProperty(Download.prototype, 'failed', {
        get()
        {
            return !!(this.currentState && this.currentState.id === 'error');
        },
        enumerable: true,
        configurable: true,
    });

    return Download;
})(Flow);

const DeleteFlow = /** @class */ (function (_super)
{
    __extends(DeleteFlow, _super);
    function DeleteFlow(gameData, basePath)
    {
        const _this = _super.call(this) || this;

        _this.data = gameData;
        _this.basePath = basePath;
        const zipDestination = `${basePath}zips/${_this.data.downloadID}/${_this.data.versionID}.zip`;
        const gameDestination = _this.unpackPath;

        if (hasCordova())
        {
            // create the download steps
            _this.add(new Waiting())
                .add(new Cleaning(zipDestination)) // delete the zip
                .add(new Cleaning(gameDestination)) // delete the zip
                .add(new Done());
        }
        else
        {
            _this.add(new Waiting())
                .add(new SleepState('CLEANING', 100))
                .add(new SleepState('CLEANING', 200))
                .add(new Done());
        }
        if (gameData.flowIndex !== undefined)
        {
            _this.set(gameData.flowIndex);
        }

        return _this;
    }
    Object.defineProperty(DeleteFlow.prototype, 'finished', {
        get()
        {
            return this.getIndex() === 4;
        },
        enumerable: true,
        configurable: true,
    });
    Object.defineProperty(DeleteFlow.prototype, 'unpackPath', {
        get()
        {
            return `${this.basePath}downloads/${this.data.downloadID}/${this.data.versionID}`;
        },
        enumerable: true,
        configurable: true,
    });
    Object.defineProperty(DeleteFlow.prototype, 'downloadID', {
        get()
        {
            return this.data.downloadID;
        },
        enumerable: true,
        configurable: true,
    });

    return DeleteFlow;
})(Flow);

/**
 * removes the any duplicate flows from an array.
 *
 * @param array - an array of download data
 */
function removeDuplicateFlows(array)
{
    const hash = {};
    const filtered = array.filter(function (flow)
    {
        if (hash[flow.downloadID])
        {
            return false;
        }
        hash[flow.downloadID] = true;

        return true;
    });

    return filtered;
}

const DownloadManager = /** @class */ (function ()
{
    function DownloadManager()
    {
        this._preferCordova = false;
        this.storage = new Storage('download_data.json');
        this.downloadMetadata = {};
        this.activeFlows = [];
        this.basePath = '';
        this.isActive = false;
    }
    /**
     * Call this to initiate the downloadManager. This must be called prior to calling any other functionality.
     *
     * @param {InitParameters} [initParams] the data object used for initiating the downloads
     * @param {Array} [initParams.incomingDownloadEntries] an array of DownloadData that the downloadmanager has access to
     * @param {string} [initParams.basePath] the base path location for file writing. Not necessary in most cordova envs (see line 58)
     * @param {boolean} [initParams.autoStart = true] setting to true will automatically resume the download process behind the scenes without the need to call `resume`
     * @param {boolean} [initParams.preferCordova = false] setting to true will download items using the `cordova-plugin-file-transfer` rather than `XMLHttpRequest`
     * @example
     * ```
     * const myDownloadData = [
     *   {
     *     downloadID: "air-show",
     *     versionID: "abcdefg",
     *     archivePath: "https://my.great.website.com/air-show/.."
     *   },
     *   {
     *     downloadID: "nature-cats-adventure",
     *     versionID: "hijklmnop",
     *     archivePath: "https://my.great.website.com/nature-cats-adventure/.."
     *   },
     * ];
     * DownloadManager.init({ incomingDownloadEntries: myDownloadData });
     * ```
     */
    DownloadManager.prototype.init = function (initParams)
    {
        let _a;
        let _b;
        let _c;
        let _d;

        return __awaiter(this, void 0, void 0, function ()
        {
            let _i;
            let _e;
            let downloadEntry;
            let savedDownloads;
            let savedDeleteFlows;
            let savedData;
            let e_1;
            let downloadID;
            let versionID;
            let recoveredDownloadData;
            let recoveredDownload;
            let deleteFlows;
            const _this = this;

            return __generator(this, function (_f)
            {
                switch (_f.label)
                {
                    case 0:
                        if (!hasCordova())
                        {
                            console.warn('DowloadManager init: cordova not found, mock api will be used');
                        }
                        if (hasCordova())
                        {
                            this.basePath = cordova.file.dataDirectory;
                        }
                        else
                        {
                            this.basePath = '../';
                        }
                        this.downloadMetadata = {};
                        if (initParams)
                        {
                            this._preferCordova = (_a = initParams.preferCordova) !== null && _a !== void 0 ? _a : this._preferCordova;
                            if (!initParams.incomingDownloadEntries)
                            {
                                console.warn('DowloadManager init: downloadMetadata array was not provided, nothing will be downloadable');
                            }
                            if (!initParams.basePath)
                            {
                                console.warn('DowloadManager init: basepath not provided, cordova.file.dataDirectory will be used');
                            }
                            if (initParams.incomingDownloadEntries)
                            {
                                for (_i = 0, _e = initParams.incomingDownloadEntries; _i < _e.length; _i++)
                                {
                                    downloadEntry = _e[_i];
                                    // check for keys
                                    if (!downloadEntry.versionID || !downloadEntry.downloadID)
                                    {
                                        throw new Error('Download entries must include a downloadID and a versionID');
                                    }
                                    // make initial entry
                                    if (!this.downloadMetadata[downloadEntry.downloadID])
                                    {
                                        this.downloadMetadata[downloadEntry.downloadID] = {
                                            activeVersion: downloadEntry.versionID,
                                            downloads: {},
                                            versions: {},
                                        };
                                    }
                                    // make version entry
                                    this.downloadMetadata[downloadEntry.downloadID].versions[downloadEntry.versionID] = {
                                        archivePath: downloadEntry.archivePath,
                                        installed: false,
                                        unpackedSize: downloadEntry.unpackedSize,
                                        updated: downloadEntry.updated,
                                    };
                                }
                            }
                        }
                        this.isActive = false;
                        this.activeFlows = [];
                        console.log('getting storage..');
                        _f.label = 1;
                    case 1:
                        _f.trys.push([1, 3, , 4]);

                        return [4 /* yield*/, this.storage.load()];
                    case 2:
                        savedData = (_f.sent());
                        // we end up using this directly so we can cast it to DownloadMetadata
                        savedDownloads = savedData === null || savedData === void 0 ? void 0 : savedData.downloads;
                        savedDeleteFlows = savedData === null || savedData === void 0 ? void 0 : savedData.deleteFlows;
                        if (savedDeleteFlows)
                        {
                            savedDeleteFlows = removeDuplicateFlows(savedDeleteFlows);
                        }

                        return [3 /* break*/, 4];
                    case 3:
                        e_1 = _f.sent();
                        throw new Error('cannot read download-state.json');
                    case 4:
                        console.log('got storage..', JSON.stringify(savedDownloads));
                        // For each Download...
                        for (downloadID in savedDownloads)
                        {
                            if (!this.downloadMetadata[downloadID])
                            {
                                // totally new entry to me!
                                this.downloadMetadata[downloadID] = savedDownloads[downloadID];
                                this.downloadMetadata[downloadID].downloads = {};
                            }
                            if (!this.downloadMetadata[downloadID].activeVersion)
                            {
                                this.downloadMetadata[downloadID].activeVersion = savedDownloads[downloadID].activeVersion;
                            }
                            // For each version of that download...
                            for (versionID in savedDownloads[downloadID].versions)
                            {
                                // any new versions, add them in
                                if (!this.downloadMetadata[downloadID].versions[versionID])
                                {
                                    this.downloadMetadata[downloadID].versions[versionID] = savedDownloads[downloadID].versions[versionID];
                                }
                                // we always re-build download entry from data
                                // as the download can then be resumed when the app is opened again
                                if ((_c = (_b = savedDownloads[downloadID]) === null || _b === void 0 ? void 0 : _b.versions[versionID]) === null || _c === void 0 ? void 0 : _c.archivePath)
                                {
                                    recoveredDownloadData = {
                                        archivePath: savedDownloads[downloadID].versions[versionID].archivePath,
                                        downloadID,
                                        flowIndex: savedDownloads[downloadID].versions[versionID].flowIndex,
                                        versionID,
                                        updated: savedDownloads[downloadID].versions[versionID].updated,
                                        unpackedSize: savedDownloads[downloadID].versions[versionID].unpackedSize,
                                    };
                                    recoveredDownload = new Download(recoveredDownloadData, this.basePath, this._preferCordova);
                                    this.downloadMetadata[downloadID].downloads[versionID] = recoveredDownload;
                                    this.downloadMetadata[downloadID].versions[versionID].installed = recoveredDownload.finished;
                                    // if the download is not finished, lets push it back into the active queue
                                    if (!recoveredDownload.finished)
                                    {
                                        this.activeFlows.push(recoveredDownload);
                                    }
                                }
                            }
                            // make sure that we order the flows so any active ones start first!
                            this.activeFlows.sort(function (a, b)
                            {
                                return b.getIndex() - a.getIndex();
                            });
                            deleteFlows = (_d = savedDeleteFlows === null || savedDeleteFlows === void 0 ? void 0 : savedDeleteFlows.map(function (deleteData)
                            {
                                return new DeleteFlow(deleteData, _this.basePath);
                            })) !== null && _d !== void 0 ? _d : [];
                            this.activeFlows = deleteFlows.concat(this.activeFlows);
                        }
                        if (!initParams || initParams.autoStart || initParams.autoStart === undefined)
                        {
                            this.resume();
                        }

                        return [2];
                }
            });
        });
    };
    /**
     * returns an array of all in progress / completed downloads.
     * @return {Array} array of in progress / completed downloads.
     */
    DownloadManager.prototype.listAllDownloads = function ()
    {
        const _list = [];

        for (const downloadEntry in this.downloadMetadata)
        {
            for (const versionEntry in this.downloadMetadata[downloadEntry].downloads)
            {
                const downloadObject = this.downloadMetadata[downloadEntry].downloads[versionEntry];

                _list.push(downloadObject);
            }
        }

        return _list;
    };
    /**
     * returns an array of one download object per download entry; specifically, the activeVersion.
     * @return {Array} array of active downloads.
     */
    DownloadManager.prototype.listActiveDownloads = function ()
    {
        const _list = [];

        for (const downloadEntry in this.downloadMetadata)
        {
            const activeVersionID = this.downloadMetadata[downloadEntry].activeVersion;
            const activeDownloadObject = this.downloadMetadata[downloadEntry].downloads[activeVersionID];

            _list.push(activeDownloadObject);
        }

        return _list;
    };
    /**
     *
     * @param {string} downloadID the downloadID of the download you would like to retreive
     * @param {string} [versionID] the versionID of the download you would like to retireve. Defaults to the active version if unspecified.
     * @return {Download | null} returns the Download if it exists, otherwise returns null
     */
    DownloadManager.prototype.getDownload = function (downloadID, versionID)
    {
        // use active
        if (!versionID)
        {
            versionID = this.downloadMetadata[downloadID].activeVersion;
        }

        return this.downloadMetadata[downloadID].downloads[versionID] || null;
    };
    /**
     * The main function! Calling this will begin the download of the game.
     * This does not happen instantly, but instead adds the game to an internal download queue
     * The downloadID must exist in the downloadMetadata list provided or the promise will be rejected.
     * The returned download is specific to the version requested and can be monitored to track progress.
     * @param {string} downloadID the downloadID of the game to be downloaded.
     * @param {string} requestedVersion the versionCode of the thing you want to be downloaded
     * @param {boolean} prioritize if true, pushes this download to the front of the queue instead of the back
     *
     * @example
     * ```
     * DownloadManager.init([
     *   {
     *     downloadID: "air-show",
     *     versionID: "abcd45578",
     *     archivePath: "https://my.great.website.com/air-show/..."
     *   }
     * ]);
     * DownloadManager.download("air-show", "abcd45578");
     * ```
     * @return {Promise<Download>} resolved with a Download object that can be monitored
     */
    DownloadManager.prototype.download = function (requestedDownloadID, requestedVersion, prioritize)
    {
        if (prioritize === void 0)
        {
            prioritize = false;
        }

        return __awaiter(this, void 0, void 0, function ()
        {
            let downloadEntry;
            let requestedDownload;
            let markInstalled;
            const _this = this;

            return __generator(this, function (_a)
            {
                switch (_a.label)
                {
                    case 0:
                        downloadEntry = this.downloadMetadata[requestedDownloadID];
                        if (!downloadEntry)
                        {
                            throw new Error(`Download ${requestedDownloadID} not found`);
                        }
                        // check if there is an existing dlObject
                        requestedDownload = downloadEntry.downloads[requestedVersion];
                        if (requestedDownload) return [3 /* break*/, 4];
                        console.log('Download not found in memory; downloading...');
                        if (!downloadEntry.versions[requestedVersion])
                        {
                            throw new Error(`Version ${requestedVersion} not found for ${requestedDownloadID}.`);
                        }
                        if (!downloadEntry.versions[requestedVersion].archivePath)
                        {
                            throw new Error(`Version ${requestedVersion} of ${requestedDownloadID} has no archivePath set.`);
                        }
                        // if not, download it!
                        requestedDownload = new Download({
                            archivePath: downloadEntry.versions[requestedVersion].archivePath || '',
                            updated: downloadEntry.versions[requestedVersion].updated || undefined,
                            unpackedSize: downloadEntry.versions[requestedVersion].unpackedSize || undefined,
                            flowIndex: 0,
                            downloadID: requestedDownloadID,
                            versionID: requestedVersion,
                        }, this.basePath, this._preferCordova);
                        markInstalled = function ()
                        {
                            _this.downloadMetadata[requestedDownloadID].versions[requestedVersion].installed = true;
                        };
                        requestedDownload.signals.onComplete.addOnce(markInstalled);
                        if (!prioritize) return [3 /* break*/, 2];

                        return [4 /* yield*/, this.pause()];
                    case 1:
                        _a.sent();
                        this.activeFlows.unshift(requestedDownload);

                        return [3 /* break*/, 3];
                    case 2:
                        this.activeFlows.push(requestedDownload);
                        _a.label = 3;
                    case 3:
                        this.downloadMetadata[requestedDownloadID].downloads[requestedVersion] = requestedDownload;
                        this.save();
                        this.resume();
                        _a.label = 4;
                    case 4: return [2 /* return*/, requestedDownload];
                }
            });
        });
    };
    /**
     * Pushes the selected download (if it exists) to the front of the queue
     * @param requestedDownloadID The downloadID of the Download you want to prioritize
     * @param requestedVersion The versionID of the Download you want to prioritize
     *
     * @example
     * ```
    * const myDownloadData = [
     *   {
     *     downloadID: "air-show",
     *     versionID: "abcdefg",
     *     archivePath: "https://my.great.website.com/air-show/.."
     *   },
     *   {
     *     downloadID: "air-show",
     *     versionID: "hijklmnop",
     *     archivePath: "https://my.great.website.com/air-show.."
     *   },
     * ];
     * DownloadManager.init({incomingDownloadEntries: myDownloadData});
     * DownloadManager.download("air-show", "abcdefg");
     * DownloadManager.download("air-show", "hijklmnop");
     * // this will pause the abc download and push the hij download to the front of the queue
     * DownloadManager.prioritize("air-show", "hijklmnop");
     * ```
     * @return {Promise} A Promise that resolves when the priority has been changed
     */
    DownloadManager.prototype.prioritize = function (requestedDownloadID, requestedVersion)
    {
        return __awaiter(this, void 0, void 0, function ()
        {
            let currentDownload;
            let found;
            let loc;
            let _i;
            let _a;
            let download;

            return __generator(this, function (_b)
            {
                switch (_b.label)
                {
                    case 0:
                        currentDownload = this.activeFlows[0];
                        if (currentDownload.data.downloadID === requestedDownloadID && currentDownload.data.versionID === requestedVersion)
                        {
                            return [2];
                        }

                        return [4 /* yield*/, this.pause()];
                    case 1:
                        _b.sent();
                        found = false;
                        loc = 0;
                        for (_i = 0, _a = this.activeFlows; _i < _a.length; _i++)
                        {
                            download = _a[_i];
                            if (download.data.downloadID === requestedDownloadID && download.data.versionID === requestedVersion)
                            {
                                found = true;
                                // remove from current
                                const down = this.activeFlows.splice(loc);

                                // put at front
                                this.activeFlows.unshift(download);
                                break;
                            }
                            loc++;
                        }
                        if (!found)
                        {
                            throw new Error(`Download ${requestedDownloadID} not found in active flows; call download with the prioritize flag set`);
                        }
                        this.save();
                        this.resume();

                        return [2];
                }
            });
        });
    };
    /**
     * Add an available download to the manager. Can be run post-init to tell the DLM about new Downloads.
     * @param {DownloadData} downloadToAdd The DownloadData for the download you want to add
     * @param {string} downloadToAdd.downloadID The ID for the download entry
     * @param {string} downloadToAdd.versionID The ID for this version of this download entry
     * @param {string} [downloadToAdd.archivePath] The location on the web from which the download is pulled
     * @example
     * ```
     * const myDownloadData = {
     *   downloadID: "air-show",
     *   versionID: "abcd45578",
     *   archivePath: "https://my.great.website.com/air-show/..."
     * }
     * await DownloadManager.addAvailableDownload(myDownloadData);
     * // can now call .download("air-show", "abcd45578");
     * ```
     * @return {Promise} a Promise that is resolved once the entry exists in the downloadMetadata
     */
    DownloadManager.prototype.addAvailableDownload = function (downloadToAdd)
    {
        if (!downloadToAdd || !downloadToAdd.downloadID || !downloadToAdd.versionID)
        {
            console.log(downloadToAdd, downloadToAdd.downloadID, downloadToAdd.versionID);
            throw new Error('Cannot add available download; download must have at least downloadID and versionID.');
        }
        if (!downloadToAdd.archivePath)
        {
            console.warn('No archivePath set. Continuing registration, but you must add an archivePath before downloading this object.');
        }
        const downloadEntry = this.downloadMetadata[downloadToAdd.downloadID];
        const newVersion = downloadToAdd.versionID;
        // 100% new dl entry, not just new version

        if (!downloadEntry)
        {
            this.downloadMetadata[downloadToAdd.downloadID] = {
                activeVersion: newVersion,
                downloads: {},
                versions: {},
            };
        }
        // existing version entry, what do?
        if (this.downloadMetadata[downloadToAdd.downloadID].versions[newVersion])
        {
            console.warn('Attempted to run addAvailableDownload on an existing entry. No action taken.');
        }
        else
        {
            const newVersionData = {
                archivePath: downloadToAdd.archivePath,
                installed: false,
                unpackedSize: downloadToAdd.unpackedSize,
                updated: downloadToAdd.updated,
            };
            // only add if a new versionData is created

            this.downloadMetadata[downloadToAdd.downloadID].versions[newVersion] = newVersionData;
        }
        console.log(`addAvailableDownload for ${downloadToAdd.downloadID} complete`);

        return Promise.resolve();
    };
    /**
     * Removes an available download from the manager.
     * @param {DeleteParams} deleteParams Object with parameter fields as follows:
     * @param {string} deleteParams.downloadID ID of the download entry
     * @param {string} [deleteParams.versionID] ID of the specific version to delete, if desired. If undefined, deletes all versions
     * @param {boolean} [deleteParams.removeMetadata] Determines whether the metadata entry is deleted. Default false. Will ALWAYS delete files if true
     * @param {boolean} [deleteParams.removeFiles] Determines whether the download's files are deleted. Defaults true
     * @example
     * ```
     * // this will delete all files for air-show, but leave the metadata intact
     * const deleteParams = {
     *   downloadID: "air-show",
     * }
     * await DownloadManager.removeAvailableDownload(myDownloadData);
     * ```
     * @example
     * // this is the same as above, if the only version is abcd45578
     * ```
     * const deleteParams = {
     *   downloadID: "air-show",
     *   versionID: "abcd45578",
     *   removeMetadata: false,
     *   removeFiles: true
     * }
     * await DownloadManager.removeAvailableDownload(myDownloadData);
     * ```
     * @return {Promise} A Promise that resolves when the delete is completed
     */
    DownloadManager.prototype.removeAvailableDownload = function (deleteParams)
    {
        return __awaiter(this, void 0, void 0, function ()
        {
            let versions;
            let _i;
            let versions_1;
            let versionID;

            return __generator(this, function (_a)
            {
                switch (_a.label)
                {
                    case 0:
                        if (!deleteParams.downloadID)
                        {
                            throw new Error();
                        }
                        if (deleteParams.removeFiles === false && deleteParams.removeMetadata === false)
                        {
                            // nothing to remove
                            console.warn('removeAvailableDownloads called with both booleans set to false; nothing deleted');

                            return [2];
                        }
                        // set defaults
                        if (!deleteParams.removeFiles === undefined)
                        {
                            deleteParams.removeFiles = true;
                        }
                        if (!deleteParams.removeMetadata === undefined)
                        {
                            deleteParams.removeMetadata = false;
                        }
                        if (deleteParams.removeMetadata && !deleteParams.removeFiles)
                        {
                            console.warn('It is not permitted to delete the metadata without deleting the associated files.');
                            console.warn('Setting removeFiles to true and continuing remove process.');
                            deleteParams.removeFiles = true;
                        }
                        if (!deleteParams.versionID) return [3 /* break*/, 2];

                        return [4 /* yield*/, this._delete(deleteParams.downloadID, deleteParams.versionID, deleteParams.removeFiles, deleteParams.removeMetadata)];
                    case 1:
                        _a.sent();

                        return [3 /* break*/, 6];
                    case 2:
                        console.warn(`Deleting all versions of ${deleteParams.downloadID}`);
                        versions = Object.keys(this.downloadMetadata[deleteParams.downloadID].versions);
                        _i = 0, versions_1 = versions;
                        _a.label = 3;
                    case 3:
                        if (!(_i < versions_1.length)) return [3 /* break*/, 6];
                        versionID = versions_1[_i];

                        return [4 /* yield*/, this._delete(deleteParams.downloadID, versionID, deleteParams.removeFiles, deleteParams.removeMetadata)];
                    case 4:
                        _a.sent();
                        _a.label = 5;
                    case 5:
                        _i++;

                        return [3 /* break*/, 3];
                    case 6: return [2];
                }
            });
        });
    };
    /**
     * Changes the activeVersion of the selected Download to the selected version
     * @param downloadID The downloadID of the Download you want to change the version of
     * @param requestedVersion The version you want to make the activeVersion
     *
     * @example
     * ```
    * const myDownloadData = [
     *   {
     *     downloadID: "air-show",
     *     versionID: "abcdefg",
     *     archivePath: "https://my.great.website.com/air-show/.."
     *   },
     *   {
     *     downloadID: "air-show",
     *     versionID: "hijklmnop",
     *     archivePath: "https://my.great.website.com/air-show.."
     *   },
     * ];
     * DownloadManager.init({incomingDownloadEntries: myDownloadData});
     * DownloadManager.download("air-show", "abcdefg");
     * DownloadManager.download("air-show", "hijklmnop");
     * // Now I want a specific version
     * DownloadManager.setActiveVersion("air-show", "hijklmnop");
     * DownloadManager.getDownload("air-show"); // returns version "hijklmnop"
     * ```
     */
    DownloadManager.prototype.setActiveVersion = function (downloadID, requestedVersion)
    {
        const downloadEntry = this.downloadMetadata[downloadID];

        if (!downloadEntry)
        {
            throw new Error(`Download ${downloadID} not found`);
        }
        if (downloadEntry.activeVersion !== requestedVersion)
        {
            const reqVersionInfo = downloadEntry.versions[requestedVersion];
            // check if the requested version exists

            if (!reqVersionInfo)
            {
                throw new Error(`Requested version ${requestedVersion} does not exist for ${downloadID}; try adding it with addAvailableDownload first`);
            }
            if (reqVersionInfo.installed === false)
            {
                throw new Error('The version you have selected has not been downloaded yet. Try calling .download() instead');
            }
            // change activeVersion
            downloadEntry.activeVersion = requestedVersion;
        }
    };
    /**
     * This function will stop downloading and delete the game from disk based on the downloadID provided.
     * This does not happen instantly, but instead adds the game to an internal deletion queue.
     * The downloadID must exist in the downloadMetadata list provided or the promise will be rejected.
     *
     * @param {string} [downloadID] the downloadID of the game to be deleted.
     * @return {Promise<void>} resolves when the element has been queued for deletion
     */
    DownloadManager.prototype._delete = function (downloadID, versionID, removeFiles, removeMetadata)
    {
        return __awaiter(this, void 0, void 0, function ()
        {
            let downloadObject;
            let deleteFlow;

            return __generator(this, function (_a)
            {
                switch (_a.label)
                {
                    case 0:
                        downloadObject = this.getDownload(downloadID, versionID);
                        if (!downloadObject) return [3 /* break*/, 2];
                        // doing this just in case it's still mid-download
                        return [4 /* yield*/, this.cancelDownload(downloadObject)];
                    case 1:
                        // doing this just in case it's still mid-download
                        _a.sent();
                        _a.label = 2;
                    case 2:
                        if (removeFiles)
                        {
                            // only need to do work here if the thing is installed
                            if (this.downloadMetadata[downloadID].versions[versionID].installed)
                            {
                                // still have to do some metadata management to make it clear we've deleted the file
                                delete this.downloadMetadata[downloadID].downloads[versionID];
                                this.downloadMetadata[downloadID].versions[versionID].installed = false;
                                // now create the delete flow
                                // make sure to reset this before passing the data!
                                downloadObject.data.flowIndex = 0;
                                deleteFlow = new DeleteFlow(downloadObject.data, this.basePath);
                                this.activeFlows.push(deleteFlow);
                            }
                        }
                        if (removeMetadata)
                        {
                            delete this.downloadMetadata[downloadID].versions[versionID];
                            // if downloaded...
                            if (this.downloadMetadata[downloadID].downloads[versionID])
                            {
                                delete this.downloadMetadata[downloadID].downloads[versionID];
                            }
                        }
                        this.save();
                        // if games are only cancelled, this this list could be null!
                        // meaning, we are no longer active
                        if (!this.activeFlows.length)
                        {
                            this.isActive = false;
                        }
                        this.resume();

                        return [2 /* return*/, Promise.resolve()];
                }
            });
        });
    };
    /*
     * Calling this resume the background download / deletion activities.
     */
    DownloadManager.prototype.resume = function ()
    {
        if (this.isActive || !this.activeFlows.length)
        {
            return;
        }
        this.next();
    };
    /**
     * Calling this pause the background download / deletion activities.
     * This can be useful when for example if you need to leave the current js context (eg going to another webpage)
     * We want to ensure that if the download manager does not exist, then we have stopped all lower level cordova plugins.
     *
     * It is recommended to pause the downloadManager before going to play a game.
     *
     * @return {Promise<void>} resolves when the download manager has successfully paused
     */
    DownloadManager.prototype.pause = function ()
    {
        return __awaiter(this, void 0, void 0, function ()
        {
            let currentDownload;

            return __generator(this, function (_a)
            {
                switch (_a.label)
                {
                    case 0:
                        if (!this.isActive || !this.activeFlows.length)
                        {
                            return [2];
                        }
                        currentDownload = this.activeFlows[0];
                        currentDownload.signals.onStateChange.remove(this.onDownloadStageChange, this);
                        currentDownload.signals.onComplete.remove(this.onDownloadComplete, this);
                        currentDownload.signals.onError.remove(this.onDownloadFail, this);

                        return [4 /* yield*/, currentDownload.pause()];
                    case 1:
                        _a.sent();
                        this.isActive = false;

                        return [2];
                }
            });
        });
    };
    DownloadManager.prototype.cancelDownload = function (download)
    {
        return __awaiter(this, void 0, void 0, function ()
        {
            let index;

            return __generator(this, function (_a)
            {
                switch (_a.label)
                {
                    case 0:
                        index = this.activeFlows.indexOf(download);
                        if (index !== -1)
                        {
                            this.activeFlows.splice(index, 1);
                            download.signals.onComplete.remove(this.onDownloadComplete, this);
                            download.signals.onError.remove(this.onDownloadFail, this);
                        }

                        return [4 /* yield*/, download.delete()];
                    case 1:
                        _a.sent();
                        this.save();

                        return [2];
                }
            });
        });
    };
    DownloadManager.prototype.save = function ()
    {
        const serialisedDownloads = {};

        for (const downloadID in this.downloadMetadata)
        {
            const downloadData = this.downloadMetadata[downloadID];

            serialisedDownloads[downloadID] = {
                activeVersion: downloadData.activeVersion,
                versions: {},
            };
            for (const versionID in downloadData.versions)
            {
                serialisedDownloads[downloadID].versions[versionID] = downloadData.versions[versionID];
                // downloads may not exist
                const download = downloadData.downloads[versionID];

                if (download)
                {
                    const stateId = download.getStateId();

                    if (stateId !== 'deleted' && stateId !== 'error')
                    {
                        serialisedDownloads[downloadID].versions[versionID].flowIndex = download.getIndex();
                    }
                }
            }
        }
        // get all the delete flows...
        const serialisedDeleteFlows = this.activeFlows.filter(function (flow)
        {
            return flow instanceof DeleteFlow;
        })
            .map(function (deleteFlow)
            {
                return __assign(__assign({}, deleteFlow.data), { flowIndex: deleteFlow.getIndex() });
            });

        this.storage.save({
            downloads: serialisedDownloads,
            deleteFlows: serialisedDeleteFlows,
        });
    };
    DownloadManager.prototype.onDownloadStageChange = function ()
    {
        // show of to the world...
        this.save();
    };
    DownloadManager.prototype.onDownloadComplete = function ()
    {
        const currentDownload = this.activeFlows.shift();

        if (!currentDownload)
        {
            throw new Error('poping no active flow..');
        }
        currentDownload.signals.onStateChange.remove(this.onDownloadStageChange, this);
        currentDownload.signals.onComplete.remove(this.onDownloadComplete, this);
        currentDownload.signals.onError.remove(this.onDownloadFail, this);
        this.next();
    };
    DownloadManager.prototype.next = function ()
    {
        const currentDownload = this.activeFlows[0];

        if (!currentDownload)
        {
            // finished!
            this.isActive = false;
            this.save();
        }
        else
        {
            this.isActive = true;
            console.log('adding signals for ', currentDownload.data.downloadID);
            currentDownload.signals.onStateChange.add(this.onDownloadStageChange, this);
            currentDownload.signals.onComplete.add(this.onDownloadComplete, this);
            currentDownload.signals.onError.add(this.onDownloadFail, this);
            currentDownload.resume();
        }
    };
    DownloadManager.prototype.onDownloadFail = function (download)
    {
        // Remove failed download object from metadata
        delete this.downloadMetadata[download.downloadID].downloads[download.data.versionID];
        // remove from flow array and kill listeners
        const currentDownload = this.activeFlows.shift();

        if (!currentDownload)
        {
            throw new Error('onDownloadFail failed to find a download to clean up');
        }
        currentDownload.signals.onStateChange.remove(this.onDownloadStageChange, this);
        currentDownload.signals.onComplete.remove(this.onDownloadComplete, this);
        currentDownload.signals.onError.remove(this.onDownloadFail, this);
        // continue
        this.save();
        this.next();
    };
    // FOR DEBUGGING ONLY
    DownloadManager.prototype.reset = function ()
    {
        return __awaiter(this, void 0, void 0, function ()
        {
            return __generator(this, function (_a)
            {
                switch (_a.label)
                {
                    case 0:
                        if (!this.storage) return [3 /* break*/, 3];
                        // console.log(this.storage)
                        return [4 /* yield*/, this.storage.clear()];
                    case 1:
                        // console.log(this.storage)
                        _a.sent();

                        return [4 /* yield*/, this.pause()];
                    case 2:
                        _a.sent();
                        // TODO need to actually delete the files!
                        this.activeFlows = [];
                        this.downloadMetadata = {};

                        return [2 /* return*/, Promise.resolve()];
                    case 3: throw new Error('Cannot reset download manager as it has not been inited yet');
                }
            });
        });
    };

    return DownloadManager;
})();

const DownloadManager$1 = new DownloadManager();

module.exports = { CalculateSize, Cleaning, Copy, DeleteFlow, Done, Download, DownloadManager: DownloadManager$1, Downloading, DownloadingCordova, Flow, Signal, SleepState, Storage, Unzipping, Waiting, removeDuplicateFlows };
