(function() {
  var plugin = new window.springroll.ContainerPlugin();

  plugin.setup = function() {
    var originalOpenPath = this.openPath;

    /**
     * An override of the default openPath behavior to append on query parameters about the current deployment environment
     */
    this.openPath = function(path, options, playOptions) {
      var queryString = buildQueryString();

      // If there's already a query string, respect that and just append on more.
      if (path.indexOf("?") >= 0) {
        path += "&" + queryString;
      } else {
        path += "?" + queryString;
      }

      // Add cordova and device properties to playOptions object.
      playOptions = playOptions || {};
      if (window.device instanceof Object) {
        playOptions.cordova = device.cordova;
        playOptions.platform = device.platform;
        playOptions.model = device.model;
      }
      
      // In the container's openPath() method, playOptions is not 
      // added to options object if object is not of type boolean.
      // To circumvent that, lets add playOptions to options now.
      if (typeof options !== 'boolean') {
        options = options || {}
        options.playOptions = playOptions;
      }

      originalOpenPath.call(this, path, options, playOptions);
    };
  };

  var buildQueryString = function() {
    var params = {};

    // Requires cordova-plugin-buildinfo
    if (window.BuildInfo instanceof Object) {
      params.packageName = BuildInfo.packageName;
      params.version = BuildInfo.version;
      params.versionCode = BuildInfo.versionCode;
    }

    if (window.device instanceof Object) {
      params.cordova = device.cordova;
      params.platform = device.platform;
      params.model = device.model;
    }

    return Object.keys(params)
      .map(function(paramName) {
        return (
          encodeURIComponent(paramName) +
          "=" +
          encodeURIComponent(params[paramName])
        );
      })
      .join("&");
  };
})();
